import { z } from "https://deno.land/x/zod@v3.22.4/mod.ts";
import { channel, log_from } from "./utils.ts";
const log = log_from(import.meta.url);

export type TargetRole = { guild_id: string; role_id: string };

/**
 * A set of rules for associating CRI groups with Discord roles.
 */
export class RuleSet {
	private rules;
	private file;
	private update_channel;

	public constructor(file: string) {
		this.rules = [] as SerializedRule[];
		this.file = file;
		this.update_channel = channel<{ updated_groups: string[] }>();
	}

	/**
	 * Reads a RuleSet from a JSON serialized file.
	 */
	public static async from_file(path: string) {
		const result = new RuleSet(path);
		const file_content = await Deno.readTextFile(path);
		const parsed = parse_rules(file_content);
		for (const { group_id, target_role, include_historical } of parsed) {
			await result.append_rule(group_id, target_role, include_historical);
		}
		result.update_channel = channel();
		return result;
	}

	/**
	 * Gets which Discord roles must be assigned to a user which is part of a given CRI group.
	 */
	public *roles_for_group(group_id: string, historical: boolean) {
		for (const rule of this.rules) {
			if (rule.group_id !== group_id) continue;
			if (!rule.include_historical && historical) continue;
			yield rule.target_role;
		}
	}

	public *roles_for_guild(target_guild_id: string) {
		for (const { group_id, target_role: { guild_id, role_id }, include_historical } of this.rules) {
			if (target_guild_id === guild_id) yield { group_id, role_id, include_historical };
		}
	}

	/**
	 * Number of managed Discord roles.
	 */
	public size() {
		return this.rules.length;
	}

	public async append_rule(group_id: string, target_role: TargetRole, include_historical: boolean) {
		for (const r of this.rules) {
			const same_group = r.group_id === group_id;
			const same_target_role = r.target_role.guild_id === target_role.guild_id &&
				r.target_role.role_id === target_role.role_id;
			if (same_group && same_target_role) return;
		}
		this.rules.push({ group_id, target_role, include_historical });
		await this.save();
		this.update_channel.send({ updated_groups: [group_id] });
	}

	public async remove_rules(group_id: string, guild_id: string) {
		const removed = [] as SerializedRule[];
		const kept = [] as SerializedRule[];
		for (const rule of this.rules) {
			const is_deleted_group = rule.group_id === group_id;
			const is_from_guild = rule.target_role.guild_id = guild_id;
			if (is_deleted_group && is_from_guild) removed.push(rule);
			else kept.push(rule);
		}
		this.rules = kept;
		await this.save();
		this.update_channel.send({ updated_groups: removed.map((r) => r.group_id) });
		return removed;
	}

	public async *receive_rule_update() {
		while (true) {
			const update = await this.update_channel.receive();
			log(`Received an update for roles`, update);
			yield update;
		}
	}

	private async save() {
		const serialized = JSON.stringify(this.rules, null, 4);
		await Deno.writeTextFile(this.file, serialized);
	}
}

type SerializedRule = ReturnType<typeof parse_rules>[0];
function parse_rules(content: string) {
	return z.array(z.object({
		group_id: z.string(),
		target_role: z.object({ guild_id: z.string(), role_id: z.string() }),
		include_historical: z.boolean().default(false),
	})).parse(JSON.parse(content));
}
